﻿using System;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Threading;

//
// *** Ensure you have the 32 bit version of mir_sdr_api.dll installed in the same directory as the rest of the source code
// or you will get a compile error. Visual Studio will try to copy the DLL to the output directory when compiling. You can use 
// the 64 bit version if you wish, but make sure the project properties are changed to 64 bit.
//

namespace StreamingTest
{
    public partial class frmMain : Form
    {
		// the following are extracted from my mir_sdr_api.dll wrapper (SDRInterface.cs) to 
		// save space, and avoid unnecessary clutter. The file is included in the zip file
		// but note the namespace is different and will have to be changed if you use the
		// file in your own project.

		public enum mir_sdr_ErrT
		{
			mir_sdr_Success				= 0,
			mir_sdr_Fail				= 1,
			mir_sdr_InvalidParam		= 2,
			mir_sdr_OutOfRange			= 3,
			mir_sdr_GainUpdateError		= 4,
			mir_sdr_RfUpdateError		= 5,
			mir_sdr_FsUpdateError		= 6,
			mir_sdr_HwError				= 7,
			mir_sdr_AliasingError		= 8,
			mir_sdr_AlreadyInitialised	= 9,
			mir_sdr_NotInitialised		= 10,
			mir_sdr_NotEnabled			= 11,
			mir_sdr_HwVerError			= 12,
			mir_sdr_OutOfMemError		= 13
		}

		public enum mir_sdr_Bw_MHzT
		{
			mir_sdr_BW_Undefined = 0,
			mir_sdr_BW_0_200 = 200,
			mir_sdr_BW_0_300 = 300,
			mir_sdr_BW_0_600 = 600,
			mir_sdr_BW_1_536 = 1536,
			mir_sdr_BW_5_000 = 5000,
			mir_sdr_BW_6_000 = 6000,
			mir_sdr_BW_7_000 = 7000,
			mir_sdr_BW_8_000 = 8000
		}

		public enum mir_sdr_If_kHzT
		{
			mir_sdr_IF_Undefined = -1,
			mir_sdr_IF_Zero  = 0,
			mir_sdr_IF_0_450 = 450,
			mir_sdr_IF_1_620 = 1620,
			mir_sdr_IF_2_048 = 2048
		}

		public enum mir_sdr_SetGrModeT
		{
			mir_sdr_USE_SET_GR = 0,
			mir_sdr_USE_SET_GR_ALT_MODE = 1,
			mir_sdr_USE_RSP_SET_GR = 2
		}

		// the following two enums are not part of the offical API, but have been added for 
		// clarity in the code, and avoids having to use "magic" numbers. There are
		// several other usefull enums declared in SDRInterface.cs

		public enum mir_sdr_DecimationT : uint
		{
			mir_sdr_DISABLED = 0,
			mir_sdr_ENABLED  = 1
		}

		public enum mir_sdr_DecimationFilterT : uint
		{
			mir_sdr_AVERAGING = 0,
			mir_sdr_HALFBAND  = 1
		}

		// DLL import declarations
		[DllImport("mir_sdr_api.dll", CallingConvention = CallingConvention.Cdecl)]
		private static extern mir_sdr_ErrT mir_sdr_StreamInit(ref int gRdB, double fsMHz, double rfMHz, mir_sdr_Bw_MHzT bwType, mir_sdr_If_kHzT ifType, int LNAstate, ref int gRdBsystem, mir_sdr_SetGrModeT setGrMode, ref int samplesPerPacket, mir_sdr_StreamCallback_t StreamCbFn, mir_sdr_GainChangeCallback_t GainChangeCbFn, IntPtr cbContext);

		[DllImport("mir_sdr_api.dll", CallingConvention = CallingConvention.Cdecl)]
		private static extern mir_sdr_ErrT mir_sdr_StreamUninit();

		[DllImport("mir_sdr_api.dll", CallingConvention = CallingConvention.Cdecl)]
		private static extern mir_sdr_ErrT mir_sdr_DecimateControl(mir_sdr_DecimationT enable, uint decimationFactor, mir_sdr_DecimationFilterT wideBandSignal);

		// streaming callback delegate declaration
		[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
		public delegate void mir_sdr_StreamCallback_t(IntPtr xi, IntPtr xq, uint firstSampleNum, int grChanged, int rfChanged, int fsChanged, uint numSamples, uint reset, IntPtr cbContext);

		// gain-change callback delegate declaration
		[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
		public delegate void mir_sdr_GainChangeCallback_t(uint gRdB, uint lnaGRdB, IntPtr cbContext);

		// streaming and gain change callback function pointers
		public mir_sdr_StreamCallback_t streamcbfuncptr;
		public mir_sdr_GainChangeCallback_t gcfuncptr;

		// global error variable
        public mir_sdr_ErrT r;

		// streaming buffers and buffer lock
		public static short[] Ibuf;
        public static short[] Qbuf;
		public object bufferLock;

		// used to indicate streaming activity
		public int numPkts = 0;
        public int numEvents = 0;

		// params for streaming startup
		public int gr;
		public double fsHz;
		public double rfMHz;
		public int LNAstate;
		public int gRdBsystem;
		public int samplesPerPacket;

		public bool SDRIsStreaming;

		// decimation
		public UInt32 decimationFactor;
		public double effectiveSampleRate;

		public int heartbeatTicks;

		public frmMain()
        {
            InitializeComponent();

			// init IQ buffering
			Ibuf = new short[16384];
			Qbuf = new short[16384];
			bufferLock = new object();

			// get initial startup params - no validation required unless you mess with the form design and leave blank or character entries
			rfMHz = Convert.ToDouble(txtRfFreq.Text);
			fsHz = Convert.ToDouble(txtSampFreq.Text);
			decimationFactor = (Convert.ToUInt32(cmbDecimation.Text));

			// calculate and display initial effective sampling rate
			effectiveSampleRate = fsHz / decimationFactor;
			lblEffectiveSampleRate.Text = effectiveSampleRate.ToString();

			// plug in the streaming and gain change callback functions
			streamcbfuncptr = new mir_sdr_StreamCallback_t(streamcbfunc);
			gcfuncptr = new mir_sdr_GainChangeCallback_t(gainChangeCallBackFunction);

			// not streaming yet
			SDRIsStreaming = false;

			// start the heartbeat timer - does nothing useful but used to show that the program has not locked up
			tmrHeartbeat.Interval = 1000;
            tmrHeartbeat.Enabled = true;
		}

		private void btnStart_Click(object sender, EventArgs e)
		{
			// set initial state
			samplesPerPacket = 0;
			gr = 30;
			gRdBsystem = 0;
			LNAstate = 0;

			// context not used - did this to stop the compiler complaining
			int dummy = 0;
			IntPtr context = (IntPtr)dummy;

			// quick and dirty way to validate RX frequency, sampling frequency and decimation factor - if entry is not numeric, or is blank, then refuse to run
			try { rfMHz = Convert.ToDouble(txtRfFreq.Text); } catch {  return; }
			try { fsHz = Convert.ToDouble(txtSampFreq.Text); } catch { return; }
			try { decimationFactor = (Convert.ToUInt32(cmbDecimation.Text)); } catch { return; }

			// calculate and display effective sampling rate
			effectiveSampleRate = fsHz / decimationFactor;
			lblEffectiveSampleRate.Text = effectiveSampleRate.ToString();

			// start streaming
			r = mir_sdr_StreamInit(ref gr, fsHz, rfMHz, mir_sdr_Bw_MHzT.mir_sdr_BW_1_536, mir_sdr_If_kHzT.mir_sdr_IF_Zero, LNAstate, ref gRdBsystem, mir_sdr_SetGrModeT.mir_sdr_USE_SET_GR, ref samplesPerPacket, streamcbfuncptr, gcfuncptr, context);

			// if streaming started successfully then set decimation factor - this must only be set AFTER streaming is started, it pays to RTFM :-(
			if (r == mir_sdr_ErrT.mir_sdr_Success)
			{
				Console.WriteLine("SDR streaming started");

				btnStart.Enabled = false;
				btnStop.Enabled = true;
				txtRfFreq.Enabled = false;
				cmbDecimation.Enabled = false;
				txtSampFreq.Enabled = false;
				numEvents = 0;
				SDRIsStreaming = true;

				Console.WriteLine("Setting decimation");

				// Warning: returns success even if decimation set to something that is not 2^n. Invalid entry is ignored and decimation is not set according to API documententation
				if (decimationFactor > 1)
					// enable decimation
					r = mir_sdr_DecimateControl(mir_sdr_DecimationT.mir_sdr_ENABLED, decimationFactor, mir_sdr_DecimationFilterT.mir_sdr_AVERAGING);
				else
					// disable decimation, decimation factor and filtering ignored
					r = mir_sdr_DecimateControl(mir_sdr_DecimationT.mir_sdr_DISABLED, decimationFactor, mir_sdr_DecimationFilterT.mir_sdr_AVERAGING);

				if (r == mir_sdr_ErrT.mir_sdr_Success)
					Console.WriteLine("Decimation set");
				else
				{
					Console.WriteLine("Problem setting decimation");
					sdr_reportError(r, 1);
				}
			}
			else
			{
				Console.WriteLine("Problem starting SDR streaming");
				sdr_reportError(r, 2);
				SDRIsStreaming = false;
			}
		}

		public void streamcbfunc(IntPtr xi, IntPtr xq, uint firstSampleNum, int grChanged, int rfChanged, int fsChanged, uint numSamples, uint reset, IntPtr cbContext)
        {
			// lock buffers and get a copy of the IQ data - use a lock because IQ processing will probably be done in another thread
			lock (bufferLock)
			{
				Marshal.Copy(xi, Ibuf, 0, (int)numSamples);
				Marshal.Copy(xq, Qbuf, 0, (int)numSamples);
			}

			//
			// IQ data now in Ibuf and Qbuf - do something with it...
			//

			// just something to show that streaming callback is working
			numPkts++;
			if (numPkts > 100)
			{
				numPkts = 0;
				numEvents++;
				Console.WriteLine(numEvents.ToString());
			}
		}

		public void gainChangeCallBackFunction(uint gRdB, uint lnaGRdB, IntPtr cbContext)
        {
            // do nothing... only testing
        }

		public void btnStop_Click(object sender, EventArgs e)
		{
			if (SDRIsStreaming)
			{
				// stop streaming
				r = mir_sdr_StreamUninit();

				if (r == mir_sdr_ErrT.mir_sdr_Success)
				{
					Console.WriteLine("SDR stopped streaming");
					btnStart.Enabled = true;
					btnStop.Enabled = false;
					txtRfFreq.Enabled = true;
					cmbDecimation.Enabled = true;
					txtSampFreq.Enabled = true;

					SDRIsStreaming = false;
				}
				else
				{
					Console.WriteLine("Problem stopping SDR streaming");
					sdr_reportError(r, 3);
				}
			}
		}

		public void tmrHeartbeat_Tick(object sender, EventArgs e)
        {
			// just to show the program has not locked up
            heartbeatTicks++;
			lblHeartbeat.Text = heartbeatTicks.ToString();
		}

		public void frmMain_FormClosing(object sender, FormClosingEventArgs e)
        {
			// ensure streaming has stopped
			if (SDRIsStreaming)
			{
				btnStop_Click(this, e);
				Thread.Sleep(1000);
			}
		}

		public mir_sdr_ErrT sdr_reportError(mir_sdr_ErrT errorNum, int context)
		{
			// writes error message to console - if no error then nothing is written. 
			// context can be used to identify where in the code the error occured.

			string errMsg = "";
			string contextTag = context.ToString();

			if (errorNum != mir_sdr_ErrT.mir_sdr_Success)
			{
				switch (errorNum)
				{
					case mir_sdr_ErrT.mir_sdr_Fail:
						errMsg = "Fail";
						break;

					case mir_sdr_ErrT.mir_sdr_InvalidParam:
						errMsg = "Invalid parameter";
						break;

					case mir_sdr_ErrT.mir_sdr_OutOfRange:
						errMsg = "Out of  range";
						break;

					case mir_sdr_ErrT.mir_sdr_GainUpdateError:
						errMsg = "Gain update error";
						break;

					case mir_sdr_ErrT.mir_sdr_RfUpdateError:
						errMsg = "RF update error";
						break;

					case mir_sdr_ErrT.mir_sdr_FsUpdateError:
						errMsg = "FS update error";
						break;

					case mir_sdr_ErrT.mir_sdr_HwError:
						errMsg = "Hardware error";
						break;

					case mir_sdr_ErrT.mir_sdr_AliasingError:
						errMsg = "Aliasing error";
						break;

					case mir_sdr_ErrT.mir_sdr_AlreadyInitialised:
						errMsg = "Already initialised";
						break;

					case mir_sdr_ErrT.mir_sdr_NotInitialised:
						errMsg = "Not initialised";
						break;

					case mir_sdr_ErrT.mir_sdr_NotEnabled:
						errMsg = "Not enabled";
						break;

					case mir_sdr_ErrT.mir_sdr_HwVerError:
						errMsg = "Hardware version error";
						break;

					case mir_sdr_ErrT.mir_sdr_OutOfMemError:
						errMsg = "Out of memory error";
						break;

					default:
						errMsg = "Unknown error";
						break;
				}

				Console.WriteLine(contextTag + "-" + errMsg);
			}

			return errorNum;
		}

	}
}

